"""
Coroutines tool box
"""
import typing
import logging
import asyncio
from asyncio.queues import Queue, QueueEmpty
from itertools import islice

import async_timeout


def timeout_wrapper(
        coro_or_func: typing.Union[typing.Coroutine, typing.Callable],
        timeout: typing.Union[float, int]
        ) -> typing.Union[typing.Coroutine, typing.Callable]:
    """Add timeout limit to coroutine or coroutine function"""
    is_coroutinefunction = asyncio.iscoroutinefunction(coro_or_func)

    async def wrapper(*args, **kwargs):
        with async_timeout.timeout(timeout):
            if is_coroutinefunction:
                return await coro_or_func(*args, **kwargs)
            else:
                return await coro_or_func

    if timeout < 0:
        return coro_or_func
    elif is_coroutinefunction:
        return wrapper
    else:
        return wrapper()


class CoroutinesPool:
    """
    Coroutines concurrent execute pool with API like standard
    "asyncio.ensure_future" and "asyncio.as_completed'
    """

    def __init__(self, loop: typing.Optional[asyncio.AbstractEventLoop] = None,
                 raise_exception: bool = True,
                 limit: int = -1, timeout: typing.Union[int, float] = -1):
        """
        :param loop: set to "asyncio.get_event_loop()" by default.
        :param raise_exception: raise coroutine`s exception inside
            (handle by asyncio) or just log it by logging
        :param limit: concurrency coroutines`s count limit,
            no limit by default.
        :param timeout: global timeout for single one coroutine, no timeout
            by default.
        """

        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop

        self._timeout = timeout
        self._limit = limit
        self._raise_exception = raise_exception
        self._queue = Queue(
            loop=self._loop)  # store coroutines waiting for running
        self._done_queue = Queue(loop=self._loop)
        self._running_count = 0  # coroutines running count, less than limit
        self._is_closed = False  # pool status
        self.logger = logging.getLogger(self.__class__.__name__)

    @property
    def loop(self):
        return self._loop

    @property
    def timeout(self) -> typing.Union[int, float]:
        return self._timeout

    @property
    def limit(self) -> int:
        return self._limit

    @property
    def raise_exception(self) -> bool:
        return self._raise_exception

    @property
    def running_count(self) -> int:
        return self._running_count

    @property
    def is_running(self) -> bool:
        return self._running_count > 0

    @property
    def status(self):
        if self._is_closed:
            status = 'closed'
        elif self.is_running:
            status = 'running'
        else:
            status = 'ready'
        return status

    def reset(self, limit: typing.Optional[int] = None,
              timeout: typing.Union[int, float, None] = None,
              raise_exception: typing.Optional[bool] = None):
        """Rest limit or timeout or raise_exception,
        it`s safe to call anytime
        """
        self._limit = self._limit if limit is None else limit
        self._timeout = self._timeout if timeout is None else timeout
        if raise_exception is not None:
            self._raise_exception = raise_exception

    def schedule_coroutine(self, coroutine: typing.Coroutine,
                           timeout: typing.Union[
                               float, int, None] = None) -> None:
        """Like "asyncio.ensure_future", it schedule coroutine in event loop
        and return immediately.

        Call "self.wait_scheduled_coroutines" make sure all coroutine has been
        done.

        :param timeout: set to "self._timeout" by default
        """
        timeout = self._timeout if timeout is None else timeout

        def _done_callback(f):
            self._running_count -= 1
            self._done_queue.put_nowait(f)
            try:
                if self._limit == -1 or self._running_count < self._limit:
                    next_coroutine = self._queue.get_nowait()
                    self._running_count += 1
                    asyncio.ensure_future(next_coroutine,
                                          loop=self._loop).add_done_callback(
                        _done_callback)
            except QueueEmpty:
                pass
            # handle exception
            exception = f.exception()
            if exception is not None:
                try:
                    raise exception
                except exception.__class__:
                    if self._raise_exception:
                        raise exception
                    else:
                        self.logger.exception(exception)

        if self._is_closed:
            self.logger.warning('This pool has be closed!')
            return

        if self._limit == -1 or self._running_count < self._limit:
            self._running_count += 1
            asyncio.ensure_future(
                timeout_wrapper(coroutine, timeout=timeout),
                loop=self._loop).add_done_callback(_done_callback)
        else:
            self._queue.put_nowait(timeout_wrapper(coroutine, timeout=timeout))

    def schedule_coroutines(self,
                            coroutines: typing.Iterable[typing.Coroutine],
                            timeout: typing.Union[
                                float, int, None] = None) -> None:
        """A short way to schedule many coroutines.

        :param timeout: set to "self._timeout" by default
        """
        for coroutine in coroutines:
            self.schedule_coroutine(coroutine, timeout=timeout)

    async def wait_scheduled_coroutines(self):
        """Wait scheduled coroutines to be done, can be called many times.
        """
        while self._running_count > 0 or self._done_queue.qsize() > 0:
            await self._done_queue.get()

    def as_completed(self, coroutines: typing.Iterable[typing.Coroutine],
                     limit: typing.Optional[int] = None,
                     timeout: typing.Union[int, float, None] = None
                     ) -> typing.Generator[typing.Coroutine, None, None]:
        """Like "asyncio.as_completed",
        run and iter coroutines out of the pool.

        :param timeout: set to "self._timeout" by default
        :param limit: set to "self._limit" by default, this "limit" is not
            shared with pool`s limit
        """
        limit = self._limit if limit is None else limit
        timeout = self._timeout if timeout is None else timeout

        coroutines = iter(coroutines)
        queue = Queue(loop=self._loop)
        todo = []

        def _done_callback(f):
            queue.put_nowait(f)
            todo.remove(f)
            try:
                nf = asyncio.ensure_future(next(coroutines))
                nf.add_done_callback(_done_callback)
                todo.append(nf)
            except StopIteration:
                pass

        async def _wait_for_one():
            f = await queue.get()
            return f.result()

        if limit <= 0:
            fs = {asyncio.ensure_future(
                timeout_wrapper(cor, timeout=timeout),
                loop=self._loop
            ) for cor in coroutines}
        else:
            fs = {asyncio.ensure_future(
                timeout_wrapper(cor, timeout=timeout),
                loop=self._loop
            ) for cor in islice(coroutines, 0, limit)}
        for f in fs:
            f.add_done_callback(_done_callback)
            todo.append(f)

        while len(todo) > 0 or queue.qsize() > 0:
            yield _wait_for_one()

    async def as_completed_with_async(
            self, coroutines: typing.Iterable[typing.Coroutine],
            limit: typing.Optional[int] = None,
            timeout: typing.Union[int, float, None] = None,
            raise_exception: typing.Optional[bool] = None,
    ) -> typing.AsyncGenerator[typing.Any, None]:
        """as_completed`s async version, can catch and log exception inside.

        :param timeout: set to "self._timeout" by default
        :param limit: set to "self._limit" by default, this "limit" is not
            shared with pool`s limit
        :param raise_exception: set to "self._raise_exception" by default
        """
        if raise_exception is None:
            raise_exception = self._raise_exception

        for coro in self.as_completed(coroutines, limit=limit,
                                      timeout=timeout):
            try:
                yield await coro
            except Exception as e:
                if raise_exception:
                    raise e
                else:
                    self.logger.exception(e)

    async def close(self):
        self._is_closed = True
        await self.wait_scheduled_coroutines()

    def __repr__(self):
        return '{:s}({:s}) with {:d} running coroutines count'.format(
            self.__class__.__name__, self.status, self._running_count)

    def __del__(self):
        if not self._is_closed:
            if self._running_count > 0 or self._done_queue.qsize() > 0:
                self.logger.error('Leave with running coroutines!')
            else:
                self.logger.warning('Leave without pool closed!')


__all__ = ['CoroutinesPool', 'timeout_wrapper']



"""Provide Ant`s Request, Response, Item and Extractor."""
from typing import Any, Optional, Iterator, Tuple, Dict, Type, Union, List, \
    DefaultDict, AnyStr, IO, Callable, Generator
from collections.abc import MutableMapping
import abc
from collections import defaultdict, namedtuple
import logging
import re

from aiohttp import ClientResponse, ClientRequest
from lxml import html
import jpath
import ujson

from .exceptions import FieldValidationError, ItemExtractError

ConnectionKey = namedtuple('ConnectionKey',
                           ['host', 'port', 'ssl', 'proxy_host', 'proxy_port'])


class Request(ClientRequest):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # store data obj
        self.data: Union[AnyStr, dict, IO, None] = kwargs.get('data', None)

    @property
    def connection_key(self):
        if self.proxy is not None:
            proxy_host = self.proxy.host
            proxy_port = self.proxy.port
        else:
            proxy_host = None
            proxy_port = None
        return ConnectionKey(self.host, self.port, self.is_ssl(), proxy_host,
                             proxy_port)


class Response(ClientResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._text = None
        self._html_element = None
        self._json = None

    def get_text(self, encoding: Optional[str] = None,
                 errors: str = 'strict') -> str:
        if self._body is None:
            raise ValueError('Read stream first')
        if self._text is None:
            if encoding is None:
                encoding = self.get_encoding()
            self._text = self._body.decode(encoding, errors=errors)
        return self._text

    @property
    def simple_text(self) -> str:
        return self.get_text(errors='ignore')

    def get_json(self, loads: Callable = ujson.loads):
        if self._json is None:
            self._json = loads(self.simple_text)
        return self._json

    @property
    def simple_json(self) -> Any:
        return self.get_json()

    @property
    def html_element(self) -> html.HtmlElement:
        if self._html_element is None:
            self._html_element = html.fromstring(self.simple_text)
        return self._html_element


class CustomNoneType:
    """Different with "None" obj ("null" in json)
    """
    pass


class IntField:
    _type = int
    storage_name = ''
    __shadow_name_prefix = '__field#'

    def __init__(self, null: bool = False, default: Any = CustomNoneType()):
        """
        :param null, is True means this field can be ignore when value have
            not been set in validation,
        :param default is None means this field have no default value
        """
        self.null = null
        self.default = default

    def __set__(self, instance: 'Item', value: Any) -> None:
        setattr(instance, self.storage_name, value)

    def __get__(self, instance: 'Item', owner: Type['Item']) -> Any:
        try:
            return getattr(instance, self.storage_name)
        except AttributeError as e:
            raise AttributeError(
                '\'{:s}\' object has no attribute \'{:s}\''.format(
                    instance.__class__.__name__,
                    self.get_name_from_shadow(self.storage_name))) from e

    def __delete__(self, instance):
        delattr(instance, self.storage_name)

    def validate(self, value: Any) -> Any:
        """:raise FieldValidationError"""
        try:
            return self._type(value)
        except (ValueError, TypeError) as e:
            raise FieldValidationError(str(e)) from e

    @classmethod
    def make_shadow_name(cls, name: str) -> str:
        if name == cls.__shadow_name_prefix:
            raise AttributeError(
                'This name: {:s} has been used internally'.format(
                    cls.__shadow_name_prefix))
        return cls.__shadow_name_prefix + name

    @classmethod
    def is_shadow_name(cls, name: str) -> bool:
        return cls.__shadow_name_prefix in name

    @classmethod
    def get_name_from_shadow(cls, name: str) -> str:
        return name.replace(cls.__shadow_name_prefix, '')


class FloatField(IntField):
    _type = float


class StringField(IntField):
    _type = str


class BytesField(IntField):
    _type = bytes


# store all item class`s field reference
_fields: DefaultDict[Type['Item'], Dict[str, IntField]] = defaultdict(dict)


class ItemMeta(abc.ABCMeta):
    def __init__(cls, name: str, bases: Tuple[type, ...],
                 attr_dict: Dict[str, Any]):
        super().__init__(name, bases, attr_dict)
        for k, v in attr_dict.items():
            if isinstance(v, IntField):
                _fields[cls][k] = v
                v.storage_name = IntField.make_shadow_name(k)
        # fetch all fields from base class
        for base_cls in bases:
            if base_cls in _fields:
                # base item class`s fields must have been set
                _fields[cls].update(_fields[base_cls])


class Item(MutableMapping, metaclass=ItemMeta):
    """Simple dict-like object with validation"""
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        # default field value
        for k, obj in _fields[self.__class__].items():
            if not isinstance(obj.default, CustomNoneType) and k not in self:
                setattr(self, k, obj.default)

    def __setattr__(self, key: str, value: Any) -> None:
        if isinstance(key, str):
            super().__setattr__(key, value)
        else:
            raise AttributeError(
                'attribute name must be string, not \'{:s}\''.format(
                    key.__class__.__name__))

    def __setitem__(self, key: str, value: Any) -> None:
        try:
            setattr(self, key, value)
        except TypeError as e:
            raise KeyError(str(e)) from e

    def __getitem__(self, key: str) -> Any:
        try:
            return getattr(self, key)
        except AttributeError as e:
            raise KeyError(str(e)) from e

    def __delitem__(self, key: str) -> None:
        try:
            delattr(self, key)
        except AttributeError as e:
            raise KeyError(str(e)) from e

    def __len__(self) -> int:
        return len(self.__dict__)

    def __iter__(self) -> Iterator[str]:
        keys = []
        for k in self.__dict__:
            if IntField.is_shadow_name(k):
                keys.append(IntField.get_name_from_shadow(k))
            else:
                keys.append(k)
        return iter(keys)

    def validate(self):
        """Validate item`s type.
        Get descriptors reference method from _fields"""
        for k, obj in _fields[self.__class__].items():
            if k in self:
                setattr(self, k, obj.validate(getattr(self, k)))
            elif not obj.null:
                raise FieldValidationError(
                    '\'{:s}.{:s}\' have no value yet'.format(
                        self.__class__.__name__, k))

    def __repr__(self):
        return '{:s}: {:s}'.format(self.__class__.__name__, str(dict(self)))

    def __str__(self):
        return '{:s}'.format(self.__class__.__name__)


Things = Union[Request, Response, Item]


class ItemExtractor:
    extract_with_take_first = 'take_first'
    extract_with_join_all = 'join_all'
    extract_with_do_nothing = 'do_nothing'

    def __init__(self, item_class: Type[Item]):
        self.item_class = item_class
        self.logger = logging.getLogger(self.__class__.__name__)
        self.paths: DefaultDict[str, List[Tuple[str, str, str]]] = \
            defaultdict(list)

    def add_xpath(self, key: str, xpath: str,
                  extract_type=extract_with_take_first):
        self.paths[key].append(('xpath', xpath, extract_type))

    def add_regex(self, key: str, pattern: str,
                  extract_type=extract_with_take_first):
        self.paths[key].append(('regex', pattern, extract_type))

    def add_jpath(self, key: str, jpath, extract_type=extract_with_take_first):
        self.paths[key].append(('jpath', jpath, extract_type))

    @staticmethod
    def _regex_search(pattern: str, data: Any) -> List[str]:
        # convert data to string
        if isinstance(data, Response):
            data = data.simple_text
        elif isinstance(data, html.HtmlElement):
            data = html.tostring(data, encoding='unicode')
        else:
            data = str(data)
        return re.findall(pattern, data)

    @staticmethod
    def _jpath_search(pattern: str, data: Any) -> List[Any]:
        # convert data to json object
        if isinstance(data, Response):
            data = data.simple_json
        elif isinstance(data, str):
            data = ujson.loads(data)
        return jpath.get_all(pattern, data)

    @staticmethod
    def _xpath_search(pattern: str, data: Any
                      ) -> List[Union[str, html.HtmlElement]]:
        if isinstance(data, Response):
            data = data.html_element
        elif not isinstance(data, html.HtmlElement):
            data = html.fromstring(str(data))
        return data.xpath(pattern)

    @staticmethod
    def extract_value(_type: str, pattern: str, data: Any,
                      extract_type=extract_with_take_first) -> Any:
        if _type == 'xpath':
            extract_value = ItemExtractor._xpath_search(pattern, data)
        elif _type == 'regex':
            extract_value = ItemExtractor._regex_search(pattern, data)
        elif _type == 'jpath':
            extract_value = ItemExtractor._jpath_search(pattern, data)
        else:
            raise ValueError('The type: {:s} not support'.format(_type))
        # handle by extract type
        if extract_type == ItemExtractor.extract_with_take_first:
            extract_value = extract_value[0]
        elif extract_type == ItemExtractor.extract_with_join_all:
            extract_value = list(filter(lambda x: isinstance(x, str),
                                        extract_value))  # join string only
            extract_value = ''.join(extract_value)
        return extract_value

    def extract(self, data: Any) -> Item:
        """Extract item from response or other data by xpath,
        jpath or regex."""
        self.logger.debug('Extract item: {:s} with path: {:s}'.format(
            self.item_class.__name__, str(self.paths)))
        item = self.item_class()
        for key, paths in self.paths.items():
            value = CustomNoneType()
            for path_type, path, extract_type in paths:
                try:
                    extract_value = ItemExtractor.extract_value(
                        path_type, path, data, extract_type=extract_type)
                except IndexError:  # IndexError is often raised
                    continue
                # check multiple path`s result
                if not isinstance(value,
                                  CustomNoneType) and value != extract_value:
                    raise ItemExtractError(
                        'Match different result: {:s} and {:s} for paths: '
                        '{:s}'.format(value, extract_value, str(paths)))
                value = extract_value
            if not isinstance(value, CustomNoneType):
                setattr(item, key, value)
        return item


class ItemNestExtractor(ItemExtractor):
    def __init__(self, root_path_type: str, root_path: str,
                 item_class: Type[Item]):
        self.root_path_type = root_path_type
        self.root_path = root_path
        super().__init__(item_class)

    def extract(self, response: Response):
        raise NotImplementedError('This method is dropped in this class')

    def extract_items(self, response: Response) -> Generator[Item, None, None]:
        base_data = self.extract_value(
            self.root_path_type, self.root_path, response,
            extract_type=self.extract_with_do_nothing)
        for data in base_data:
            yield super().extract(data)


__all__ = ['Request', 'Response', 'Item', 'ItemExtractor', 'ItemNestExtractor',
           'Things'] + \
          [var for var in vars().keys() if 'Field' in var]



"""Utilities box"""
from typing import Any, List, Type
import logging

from .ant import Ant
from .things import ItemExtractor

__all__ = ['extract_value_by_xpath', 'extract_value_by_jpath',
           'extract_value_by_regex', 'ExceptionFilter']


class CliAnt(Ant):
    async def run(self):
        pass


def extract_value(_type: str, path: str, data: Any,
                  extract_type: str = ItemExtractor.extract_with_take_first,
                  ignore_exception: bool = True,
                  default: Any = None) -> Any:
    try:
        return ItemExtractor.extract_value(_type, path, data,
                                           extract_type=extract_type)
    except Exception as e:
        if ignore_exception:
            return default
        else:
            raise e


def extract_value_by_xpath(
        path: str, data: Any,
        extract_type: str = ItemExtractor.extract_with_take_first,
        ignore_exception: bool = True,
        default: Any = None) -> Any:
    return extract_value('xpath', path, data, extract_type=extract_type,
                         ignore_exception=ignore_exception,
                         default=default)


def extract_value_by_jpath(
        path: str, data: Any,
        extract_type: str = ItemExtractor.extract_with_take_first,
        ignore_exception: bool = True,
        default: Any = None) -> Any:
    return extract_value('jpath', path, data, extract_type=extract_type,
                         ignore_exception=ignore_exception,
                         default=default)


def extract_value_by_regex(
        path: str, data: Any,
        extract_type: str = ItemExtractor.extract_with_take_first,
        ignore_exception: bool = True,
        default: Any = None) -> Any:
    return extract_value('regex', path, data, extract_type=extract_type,
                         ignore_exception=ignore_exception,
                         default=default)


class ExceptionFilter(logging.Filter):
    """A exception log filter class for logging.
    """

    def __init__(self, exceptions: List[Type[Exception]], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.exceptions = exceptions

    def filter(self, record):
        if record.exc_info:
            for e in self.exceptions:
                if record.exc_info[0] is e:
                    return False
        return True



from typing import Optional, List, Union, Dict, Callable, AnyStr, IO, \
    DefaultDict
import asyncio
import abc
import itertools
import logging
import time
import random
from collections import defaultdict

import aiohttp
from aiohttp.client import DEFAULT_TIMEOUT
from aiohttp import ClientSession
import async_timeout
from yarl import URL
from tenacity import retry
from tenacity.retry import retry_if_result, retry_if_exception_type
from tenacity.wait import wait_fixed
from tenacity.stop import stop_after_attempt

from .pipelines import Pipeline
from .things import Request, Response, Item, Things
from .coroutine_pool import CoroutinesPool, timeout_wrapper
from .exceptions import ThingDropped

__all__ = ['Ant']


class Ant(abc.ABC):
    response_pipelines: List[Pipeline] = []
    request_pipelines: List[Pipeline] = []
    item_pipelines: List[Pipeline] = []
    request_cls = Request
    response_cls = Response
    request_timeout = DEFAULT_TIMEOUT
    request_retries = 3
    request_retry_delay = 5
    request_proxies: List[Union[str, URL]] = []
    request_max_redirects = 10
    request_allow_redirects = True
    response_in_stream = False
    connection_limit = 100  # see "TCPConnector" in "aiohttp"
    connection_limit_per_host = 0
    pool_limit = 100
    pool_timeout = DEFAULT_TIMEOUT
    pool_raise_exception = False

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        # report var
        self._reports: DefaultDict[str, List[int, int]] = defaultdict(
            lambda: [0, 0])
        self._drop_reports: DefaultDict[str, List[int, int]] = defaultdict(
            lambda: [0, 0])
        self._start_time = time.time()
        self._last_time = self._start_time
        self._report_slot = 60  # report once after one minute by default
        self._session: aiohttp.ClientSession = self.make_session()
        self.pool = CoroutinesPool(limit=self.pool_limit,
                                   timeout=self.pool_timeout,
                                   raise_exception=self.pool_raise_exception)

    @property
    def name(self):
        return self.__class__.__name__

    async def request(self, url: Union[str, URL], method: str = 'GET',
                      params: Optional[dict] = None,
                      headers: Optional[dict] = None,
                      cookies: Optional[dict] = None,
                      data: Optional[Union[AnyStr, Dict, IO]] = None,
                      proxy: Optional[Union[str, URL]] = None,
                      timeout: Optional[Union[int, float]] = None,
                      retries: Optional[int] = None
                      ) -> Response:
        if not isinstance(url, URL):
            url = URL(url)
        if proxy and not isinstance(proxy, URL):
            proxy = URL(proxy)
        if timeout is None:
            timeout = self.request_timeout
        if retries is None:
            retries = self.request_retries

        req = self.request_cls(method, url, params=params, headers=headers,
                               cookies=cookies, data=data, proxy=proxy)
        req = await self._handle_thing_with_pipelines(req,
                                                      self.request_pipelines,
                                                      timeout=timeout)
        self.report(req)

        request_function = timeout_wrapper(self._request, timeout=timeout)
        if retries > 0:
            res = await self.make_retry_decorator(retries,
                                                  self.request_retry_delay)(
                request_function)(req)
        else:
            res = await request_function(req)

        res = await self._handle_thing_with_pipelines(res,
                                                      self.response_pipelines,
                                                      timeout=timeout)
        self.report(res)
        return res

    async def get(self, url: Union[str, URL], params: Optional[dict] = None,
                  headers: Optional[dict] = None,
                  cookies: Optional[dict] = None,
                  data: Optional[Union[AnyStr, Dict, IO]] = None,
                  proxy: Optional[Union[str, URL]] = None,
                  timeout: Optional[Union[int, float]] = None,
                  retries: Optional[int] = None) -> Response:
        kwargs = locals()
        kwargs.pop('self')
        return await self.request(method='GET', **kwargs)

    async def post(self, url: Union[str, URL], params: Optional[dict] = None,
                   headers: Optional[dict] = None,
                   cookies: Optional[dict] = None,
                   data: Optional[Union[AnyStr, Dict, IO]] = None,
                   proxy: Optional[Union[str, URL]] = None,
                   timeout: Optional[Union[int, float]] = None,
                   retries: Optional[int] = None) -> Response:
        kwargs = locals()
        kwargs.pop('self')
        return await self.request(method='POST', **kwargs)

    async def put(self, url: Union[str, URL], params: Optional[dict] = None,
                  headers: Optional[dict] = None,
                  cookies: Optional[dict] = None,
                  data: Optional[Union[AnyStr, Dict, IO]] = None,
                  proxy: Optional[Union[str, URL]] = None,
                  timeout: Optional[Union[int, float]] = None,
                  retries: Optional[int] = None) -> Response:
        kwargs = locals()
        kwargs.pop('self')
        return await self.request(method='PUT', **kwargs)

    async def patch(self, url: Union[str, URL], params: Optional[dict] = None,
                    headers: Optional[dict] = None,
                    cookies: Optional[dict] = None,
                    data: Optional[Union[AnyStr, Dict, IO]] = None,
                    proxy: Optional[Union[str, URL]] = None,
                    timeout: Optional[Union[int, float]] = None,
                    retries: Optional[int] = None) -> Response:
        kwargs = locals()
        kwargs.pop('self')
        return await self.request(method='PATCH', **kwargs)

    async def delete(self, url: Union[str, URL], params: Optional[dict] = None,
                     headers: Optional[dict] = None,
                     cookies: Optional[dict] = None,
                     data: Optional[Union[AnyStr, Dict, IO]] = None,
                     proxy: Optional[Union[str, URL]] = None,
                     timeout: Optional[Union[int, float]] = None,
                     retries: Optional[int] = None) -> Response:
        kwargs = locals()
        kwargs.pop('self')
        return await self.request(method='DELETE', **kwargs)

    async def head(self, url: Union[str, URL], params: Optional[dict] = None,
                   headers: Optional[dict] = None,
                   cookies: Optional[dict] = None,
                   data: Optional[Union[AnyStr, Dict, IO]] = None,
                   proxy: Optional[Union[str, URL]] = None,
                   timeout: Optional[Union[int, float]] = None,
                   retries: Optional[int] = None) -> Response:
        kwargs = locals()
        kwargs.pop('self')
        return await self.request(method='HEAD', **kwargs)

    async def collect(self, item: Item) -> None:
        self.logger.debug('Collect item: ' + str(item))
        await self._handle_thing_with_pipelines(item, self.item_pipelines)
        self.report(item)

    async def open(self) -> None:
        self.logger.info('Opening')
        for pipeline in itertools.chain(self.item_pipelines,
                                        self.response_pipelines,
                                        self.request_pipelines):
            try:
                obj = pipeline.on_spider_open()
                if asyncio.iscoroutine(obj):
                    await obj
            except Exception as e:
                self.logger.exception(
                    'Open pipelines with ' + e.__class__.__name__)

    async def close(self) -> None:
        for pipeline in itertools.chain(self.item_pipelines,
                                        self.response_pipelines,
                                        self.request_pipelines):
            try:
                obj = pipeline.on_spider_close()
                if asyncio.iscoroutine(obj):
                    await obj
            except Exception as e:
                self.logger.exception(
                    'Close pipelines with ' + e.__class__.__name__)

        await self._session.close()
        await self.pool.close()

        self.logger.info('Closed')

    @abc.abstractmethod
    async def run(self) -> None:
        """App custom entrance"""

    async def main(self) -> None:
        await self.open()
        try:
            await self.run()
        except Exception as e:
            self.logger.exception(
                'Run ant run`s coroutine with ' + e.__class__.__name__)
        # wait scheduled coroutines before "self.close" coroutine running
        await self.pool.wait_scheduled_coroutines()
        await self.close()
        # total report
        for name, counts in self._reports.items():
            self.logger.info('Get {:d} {:s} in total'.format(counts[1], name))
        for name, counts in self._drop_reports.items():
            self.logger.info('Drop {:d} {:s} in total'.format(counts[1], name))
        self.logger.info(
            'Run {:s} in {:f} seconds'.format(self.__class__.__name__,
                                              time.time() - self._start_time))

    @staticmethod
    def make_retry_decorator(retries: int, delay: float
                             ) -> Callable[[Callable], Callable]:
        return retry(wait=wait_fixed(delay),
                     retry=(retry_if_result(lambda res: res.status >= 500) |
                            retry_if_exception_type()),
                     stop=stop_after_attempt(retries + 1))

    def make_session(self) -> ClientSession:
        """Create aiohttp`s ClientSession"""
        return ClientSession(
            response_class=self.response_cls, request_class=self.request_cls,
            connector=aiohttp.TCPConnector(
                limit=self.connection_limit,
                enable_cleanup_closed=True,
                limit_per_host=self.connection_limit_per_host)
        )

    def get_proxy(self) -> Optional[URL]:
        """Chose a proxy, default by random"""
        try:
            return URL(random.choice(self.request_proxies))
        except IndexError:
            return None

    async def _handle_thing_with_pipelines(self, thing: Things,
                                           pipelines: List[Pipeline],
                                           timeout=DEFAULT_TIMEOUT) -> Things:
        """Process thing one by one, break the process chain when get
        exception.

        :raise ThingDropped"""
        self.logger.debug('Process thing: ' + str(thing))
        raw_thing = thing
        for pipeline in pipelines:
            try:
                # TODO: timeout limit?
                # See https://github.com/glenfant/stopit#id15
                thing = pipeline.process(thing)
                if asyncio.iscoroutine(thing):
                    with async_timeout.timeout(timeout):
                        thing = await thing
            except Exception as e:
                if isinstance(e, ThingDropped):
                    self.report(raw_thing, dropped=True)
                raise e
        return thing

    async def _request(self, req: Request) -> Response:
        if req.proxy is None:
            proxy = self.get_proxy()
        else:
            proxy = req.proxy
        # cookies in headers, params in url
        kwargs = dict(method=req.method, url=req.url, headers=req.headers,
                      data=req.data)
        kwargs['proxy'] = proxy
        kwargs['max_redirects'] = self.request_max_redirects
        kwargs['allow_redirects'] = self.request_allow_redirects

        # proxy auth not work in one session with many requests,
        # add auth header to fix it
        if proxy is not None:
            if proxy.scheme == 'http' and proxy.user is not None:
                kwargs['headers'][aiohttp.hdrs.PROXY_AUTHORIZATION] = \
                    aiohttp.BasicAuth.from_url(proxy).encode()

        response = await self._session._request(**kwargs)
        if not self.response_in_stream:
            await response.read()
            response.close()
            await response.wait_for_close()
        return response

    def report(self, thing: Things, dropped: bool = False) -> None:
        now_time = time.time()
        if now_time - self._last_time > self._report_slot:
            self._last_time = now_time
            for name, counts in self._reports.items():
                count = counts[1] - counts[0]
                counts[0] = counts[1]
                self.logger.info(
                    'Get {:d} {:s} in total with {:d}/{:d}s rate'.format(
                        counts[1], name, count, self._report_slot))
            for name, counts in self._drop_reports.items():
                count = counts[1] - counts[0]
                counts[0] = counts[1]
                self.logger.info(
                    'Drop {:d} {:s} in total with {:d}/{:d} rate'.format(
                        counts[1], name, count, self._report_slot))
        report_type = thing.__class__.__name__
        if dropped:
            reports = self._drop_reports
        else:
            reports = self._reports
        counts = reports[report_type]
        counts[1] += 1



from .ant import *  # noqa
from .things import *  # noqa
from .pipelines import *  # noqa
from .coroutine_pool import *  # noqa
from .exceptions import *  # noqa
from .utils import *  # noqa
from .utils import CliAnt  # noqa

__version__ = '0.34.1'

__all__ = (ant.__all__ + pipelines.__all__ + exceptions.__all__ +  # noqa
           coroutine_pool.__all__ + things.__all__ + utils.__all__)  # noqa
